"""
This module contains the classes that write metrics to their final endpoint.
"""
import socket
import threading

#pylint: disable=too-many-arguments
class MetricsWriter(object):
    """
    Writer for managing the socket connection to metrics receiver.
    """

    def __init__(self, host, port, dry_run=False):
        super(MetricsWriter, self).__init__()
        self.is_dry_run = dry_run
        self.host = host
        self.port = port
        self.sock = None

    def transmit_metric(self, name, value, timestamp, source, point_tags):
        """
        Transmit metric to the proxy.

        Arguments:
        name - the metric name (a.b.c)
        value - the numeric value for this metric
        timestamp - the timestamp for this metric
        source - this metric's host or source name
        point_tags - dictionary of key/value pairs
        """

        line = self._generate_line(name, value, timestamp, source, point_tags)
        if self.is_dry_run:
            thread_id = hex(threading.current_thread().ident)
            # command line mode you want to see this in stdout
            # in daemon mode you want to see this in a log
            # TODO: revisit and write to log when daemonized?
            print '[{} {}:{}] {}'.format(thread_id, self.host, self.port, line)

        else:
            self.sock.sendall('%s\n' % line)

    def _generate_line(self, name, value, timestamp, source, point_tags):
        """
        This should be overridden by the derived classes.  Generates the
        metric put line for the writer.

        Arguments:
        name - the metric name (a.b.c)
        value - the numeric value for this metric
        timestamp - the timestamp for this metric
        source - this metric's host or source name
        point_tags - dictionary of key/value pairs
        """

        pass

    def start(self):
        """
        Connect and open the socket
        """

        if not self.is_dry_run:
            self.sock = socket.socket()
            self.sock.settimeout(10.0)
            self.sock.connect((self.host, self.port))

    def stop(self):
        """
        Stop and shutdown the open socket
        """

        if self.sock is not None and not self.is_dry_run:
            self.sock.shutdown(socket.SHUT_RDWR)
            self.sock.close()

    def __enter__(self):
        """
        Python with block enter routine
        """

        self.start()
        return self

    def __exit__(self, the_type, value, traceback):
        """
        Python with block exit routine
        """

        self.stop()


class WavefrontMetricsWriter(MetricsWriter):
    """
    This is the metrics writer for the Wavefront proxy format.
    """

    def __init__(self, host, port, dry_run=False):
        super(WavefrontMetricsWriter, self).__init__(host, port, dry_run)

    def _generate_line(self, name, value, timestamp, source, point_tags):
        """
        Generates the line in the Wavefront proxy format.
        """
        line = '{} {} {} source="{}"'.format(
            name, value, long(timestamp), source)
        if point_tags is not None:
            for tag_key, tag_value in point_tags.iteritems():
                line = line + ' "{}"="{}"'.format(tag_key, tag_value)
        return line

class OpenTSDBMetricsWriter(MetricsWriter):
    """
    This is the metrics writer for the OpenTSDB format.
    """

    def __init__(self, host, port, dry_run=False):
        super(OpenTSDBMetricsWriter, self).__init__(host, port, dry_run)

    def _generate_line(self, name, value, timestamp, source, point_tags):
        """
        Generates the line in the OpenTSDB format.
        put <metric> <timestamp> <value> <tagk1=tagv1[ tagk2=tagv2 ...]
        """
        line = 'put {} {} {} host="{}"'.format(name, timestamp, value, source)
        if point_tags is not None:
            for tag_key, tag_value in point_tags.iteritems():
                line = line + ' "{}"="{}"'.format(tag_key, tag_value)
        return line



"""
Utility functions
"""

import ConfigParser
import csv
import datetime
import os.path
import re
import signal
import sys
import threading
import traceback
import urllib

import dateutil
import dateutil.tz

EPOCH = (datetime.datetime.utcfromtimestamp(0)
         .replace(tzinfo=dateutil.tz.tzutc()))
def unix_time_seconds(date_in):
    """
    Convert a datetime into unix epoch seconds
    Arguments:
    date_in - the datetime object to convert. This must have a tz = UTC
    """
    if not date_in.tzinfo or date_in.tzinfo != dateutil.tz.tzutc():
        date_in = date_in.replace(tzinfo=dateutil.tz.tzutc())
    return (date_in - EPOCH).total_seconds()

def urlencode_utf8(params):
    """
    Encode with utf8 characters.
    See: http://stackoverflow.com/a/8152242
    """
    if hasattr(params, 'items'):
        params = params.items()
    encoded = []
    for key, value in params:
        key = urllib.quote_plus(key.encode('utf8'), safe='/')
        if isinstance(value, list):
            for item in value:
                if isinstance(item, basestring):
                    item = urllib.quote_plus(item.encode('utf8'), safe='/')
                encoded.append('%s=%s' % (key, item))
            continue

        if isinstance(value, basestring):
            value = urllib.quote_plus(value.encode('utf8'), safe='/')

        encoded.append('%s=%s' % (key, value))

    return '&'.join(encoded)

class Configuration(object):
    """
    Base class for configurations that read from an INI file
    """

    def __init__(self, config_file_path, create_if_not_exist=False):
        super(Configuration, self).__init__()
        if not os.path.exists(config_file_path) and not create_if_not_exist:
            raise ValueError('Configuration file %s does not exist' %
                             (config_file_path))
        self.config_file_path = config_file_path
        self.config = ConfigParser.ConfigParser()
        self.config.read(config_file_path)

    def has_section(self, section):
        """
        Checks to see if the given section exists
        """
        try:
            self.config.get(section, "test")
        except ConfigParser.NoOptionError:
            return True
        except ConfigParser.NoSectionError:
            return False

    def get(self, section, key, default_value, default_section=None):
        """
        Gets a value from the configuration and returns the default if the
        section or key does not exist.

        Arguments:
        section - the section name
        key - the key in the section to retrieve
        default_value - the default value to return when section/key not found
        """

        try:
            value = self.config.get(section, key)
        except ConfigParser.NoOptionError:
            value = None
        except ConfigParser.NoSectionError:
            value = None


        if value is None:
            if default_section:
                return self.get(default_section, key, default_value, None)
            else:
                return default_value
        else:
            return value

    def getdate(self, section, key, default_value, default_section=None):
        """
        Gets a value from the configuration and returns the default if the
        section or key does not exist.  Assumes the value is stored as a
        string that is parseable by dateutil.parser.parse()

        Arguments:
        section - the section name
        key - the key in the section to retrieve
        default_value - the default value to return when section/key not found
        """

        value = self.get(section, key, None)
        if value:
            return (dateutil.parser.parse(value)
                    .replace(microsecond=0, tzinfo=dateutil.tz.tzutc()))
        elif default_section:
            return self.getdate(default_section, key, default_value, None)
        else:
            return default_value

    def getboolean(self, section, key, default_value, default_section=None):
        """
        Gets a value from the configuration and returns the default if the
        section or key does not exist.

        Arguments:
        section - the section name
        key - the key in the section to retrieve
        default_value - the default value to return when section/key not found
        """

        try:
            value = self.config.getboolean(section, key)
        except ConfigParser.NoOptionError:
            value = None
        except ConfigParser.NoSectionError:
            value = None

        if value is None:
            if default_section:
                return self.getboolean(
                    default_section, key, default_value, None)
            else:
                return default_value
        else:
            return value

    def getlist(self, section, key, default_value, default_section=None,
                delimiter=',', trim=False):
        """
        Gets a value from the configuration and returns the default if the
        section or key does not exist.  Value is assumed to be comma-separated
        list of values.  Will return a list split by ','.

        Arguments:
        section - the section name
        key - the key in the section to retrieve
        default_value - the default value to return when section/key not found
                        (assumed to be a list; not a string)
        trim - trim all items in the list
        """

        try:
            value = self.config.get(section, key)
            if value:
                value = value.split(delimiter)
        except ConfigParser.NoOptionError:
            value = None
        except ConfigParser.NoSectionError:
            value = None

        if value is None:
            if default_section:
                value = self.getlist(default_section, key, default_value, None)
            else:
                value = default_value

        if trim:
            return map(str.strip, value)
        else:
            return value

    def set(self, section, key, value, create_section=True):
        """
        Sets the value in the given section.  Creates the section if it does
        not exist.
        Arguments:
        section - the section name
        key - the key in the section to set
        value - the value to set
        create_section - create section if it does not exist?
        """

        try:
            self.config.set(section, key, value)
        except ConfigParser.NoSectionError as nosection:
            if create_section:
                self.config.add_section(section)
                self.config.set(section, key, value)
            else:
                raise nosection

    def save(self):
        """
        Save the current configuration to disk.
        """

        with open(self.config_file_path, 'w') as configfile:
            self.config.write(configfile)

def sanitize_name(_name, replace_map=None):
    """
    Replaces characters that are not supported
    default list:
    '.'  => _
    '//' => .
    '/'  => .
    '*'  => all
    r[^a-zA-Z-_0-9.] => _

    Arguments:
    _name - the name to sanitize

    Returns:
    Sanitized name
    """

    if not replace_map:
        replace_map = [
            {'*': 'all'},
            {'.': '_'},
            {'|': '.'},
            {'//': '.'},
            {'/': '.'}
        ]

    # see http://stackoverflow.com/a/27086669 for details on performance
    # of various methods of doing this
    name = _name.lower()
    for items in replace_map:
        for search, replace in items.iteritems():
            name = name.replace(search, replace)
    name = re.sub(r'[^a-z\-_0-9\.]', '_', name)
    return name

# Mapping for product names found in the CSV file to the metric prefix name
PRODUCT_NAME_TO_PREFIX = {
    'AmazonCloudWatch': 'cloudwatch',
    'Amazon DynamoDB': 'dynamodb',
    'Amazon Elastic Compute Cloud': 'ec2',
    'Amazon Elastic File System': 'efs',
    'Amazon Route 53': 'route53',
    'Amazon Simple Email Service': 'ses',
    'Amazon Simple Notification Service': 'sns',
    'Amazon Simple Queue Service': 'sqs',
    'Amazon Simple Storage Service': 's3',
    'Amazon Virtual Private Cloud': 'vpc',
    'AWS CloudHSM': 'cloudhsm',
    'AWS CloudTrail': 'cloudtrail',
    'AWS Config': 'config',
    'AWS Direct Connect': 'directconnect',
    'AWS Key Management Service': 'kms'
}

def get_aws_product_short_name(product):
    """
    Given a product name (e.g., "AWS CloudTrail"), return the metric
    name prefix for it (e.g., "cloudtrail")
    """

    if product in PRODUCT_NAME_TO_PREFIX:
        return PRODUCT_NAME_TO_PREFIX[product]
    else:
        return product.replace(' ', '')

#pylint: disable=too-few-public-methods
class LockedIterator(object):
    """
    thread-safe iterator
    """

    def __init__(self, iterator):
        self.lock = threading.Lock()
        self.iterator = iterator.__iter__()

    def __iter__(self):
        return self

    def next(self):
        """
        Get the next item in the iterator.
        Returns:
        The next item in the iteration
        Throws:
        StopIterator if next not found
        """

        self.lock.acquire()
        try:
            return self.iterator.next()
        finally:
            self.lock.release()

CANCEL_WORKERS_EVENT = threading.Event()
def parallel_process_and_wait(iterator, workers, logger=None):
    """
    Process an iterator of function pointers in parallel using the number
    of worker threads provided in the "workers" argument.
    Work is handed off to the 'worker' function in each new thread.
    Will wait for all threads to complete before returning

    Arguments:
    iterator - an iterable list of items to be passed to the worker routine
               each item should be a tuple of (function, (args))
    workers - number of workers (threads)
    """
    locked_iterator = LockedIterator(iterator)

    # start the threads
    group = []
    for _ in range(workers):
        thread = threading.Thread(target=worker,
                                  args=(locked_iterator, logger))
        thread.daemon = True
        thread.start()
        group.append(thread)

    # wait for all threads to finish
    # this while loop is here with a .join(timeout) because sometimes
    # a few threads seem to get "stuck" so we added a way to get debug
    # information every 60 seconds.
    iterations = 0
    active = 1
    while active and group and not CANCEL_WORKERS_EVENT.is_set():
        iterations = iterations + 1
        active = 0
        for thread in group:
            thread.join(60.0)
            if thread.is_alive():
                active = active + 1

        if logger and active:
            logger.debug('%d active thread(s) of %d total threads remaining',
                         active, len(group))
            if iterations % 5 == 0:
                dump_stack_traces(logger)

#pylint: disable=bare-except
def worker(locked_iterator, logger=None):
    """
    Worker for each thread created in parallel_process_and_wait()
    Arguments:
    locked_iterator - LockedIterator (thread-safe) iterator
    logger - optional logger object
    """

    while not CANCEL_WORKERS_EVENT.is_set():
        try:
            call_details = locked_iterator.next()
            call_details[0](*call_details[1])

        except StopIteration:
            break

        except:
            if logger:
                logger.exception('Failed to run thread worker')

            break

#pylint: disable=unused-argument
def script_debug(signalnum, frame):
    """
    Dump stack traces
    """

    dump_stack_traces(None)

#pylint: disable=unused-argument
def interrupt_signal_handler(signalnum, frame):
    """
    Function that gets called when SIGINT signal is sent
    """

    print 'Stopping running threads ...'
    # set the event so all worker threads will know to stop
    CANCEL_WORKERS_EVENT.set()

def setup_signal_handlers(logger):
    """
    Registers handlers for SIGINT
    """

    signal.signal(signal.SIGINT, interrupt_signal_handler)
    signal.signal(signal.SIGTERM, interrupt_signal_handler)
    signal.signal(signal.SIGUSR1, script_debug)
    if logger:
        logger.info('Signal handlers attached')

#pylint: disable=protected-access
def dump_stack_traces(logger=None):
    """
    Prints stack traces of all threads
    """

    out = []
    out.append('Threads: %d\n' % (threading.active_count()))
    for thread_id, stack in sys._current_frames().items():
        out.append('\n# Thread %s:' % thread_id)
        for filename, lineno, name, line in traceback.extract_stack(stack):
            out.append('File: "%s", line %d, in %s' % (filename, lineno, name))
            if line:
                out.append("  %s" % (line.strip()))

    if logger:
        logger.info('STACK TRACE:\n%s', '\n'.join(out))
    else:
        print 'STACK TRACE:\n%s' % ('\n'.join(out))

def hashfile(file_path, hasher, blocksize=65536):
    """
    See: http://stackoverflow.com/a/3431835
    """

    with open(file_path, 'r') as afile:
        buf = afile.read(blocksize)
        while len(buf) > 0:
            hasher.update(buf)
            buf = afile.read(blocksize)
        return hasher.hexdigest()

class CsvFileRow(object):
    """
    Row from CsvFile
    """

    def __init__(self, csvfile, row):
        """
        """
        self.csvfile = csvfile
        self.row = row

    def __str__(self):
        return str(self.row)

    def __repr__(self):
        return repr(self.row)

    def __getitem__(self, name):
        if name not in self.csvfile.header_key_to_index:
            raise ValueError('%s not in %s' %
                             (name, str(self.csvfile.header_key_to_index)))
        index = self.csvfile.header_key_to_index[name]
        return self.row[index]

class CsvFile(object):
    """
    Simple interface to the csv.reader
    """

    def __init__(self, reader, header_row_index=1):
        """
        Construct this instance.
        Arguments:
        reader - see csv.reader() csvfile parameter definition
        """

        self.csvreader = csv.reader(reader)

        # build map for header name to index
        index = 0
        while index < header_row_index:
            header_row = self.csvreader.next()
            index = index + 1

        self.header_key_to_index = {}
        index = 0
        for name in header_row:
            self.header_key_to_index[name] = index
            index = index + 1

    def __iter__(self):
        return self

    #pylint: disable=missing-docstring
    def next(self):
        return CsvFileRow(self, self.csvreader.next())



#!/usr/bin/env python
"""
This script pulls metrics from App Dynamics.  See the README for
more details on configuration.
This script relies on the AppDynamicsREST package available from pip.
(https://github.com/tradel/AppDynamicsREST)
"""

import ConfigParser
import datetime
import numbers
import re
import sys
import time

import logging.config
import dateutil.parser

from appd.request import AppDynamicsClient
from wavefront import command
from wavefront import utils
from wavefront.metrics_writer import WavefrontMetricsWriter

# default location for the configuration file.
DEFAULT_CONFIG_FILE_PATH = '/opt/wavefront/etc/wavefront-collector-appd.conf'

#pylint: disable=too-many-instance-attributes
class AppDPluginConfiguration(command.CommandConfiguration):
    """
    Stores the configuration for this plugin
    """

    #pylint: disable=too-many-statements
    def __init__(self, config_file_path):
        super(AppDPluginConfiguration, self).__init__(
            config_file_path=config_file_path)

        self.api_url = self.get('api', 'controller_url', None)
        self.api_username = self.get('api', 'username', None)
        self.api_password = self.get('api', 'password', None)
        self.api_account = self.get('api', 'account', None)
        self.api_debug = self.getboolean('api', 'debug', False)

        self.fields = self.getlist('filter', 'names', [])
        self.fields_whitelist_regex = self.getlist('filter', 'whitelist_regex', [])
        self.fields_whitelist_regex_compiled = []
        for regex in self.fields_whitelist_regex:
            self.fields_whitelist_regex_compiled.append(re.compile(regex))
        self.fields_blacklist_regex = self.getlist(
            'filter', 'blacklist_regex', [])
        self.fields_blacklist_regex_compiled = []
        for regex in self.fields_blacklist_regex:
            self.fields_blacklist_regex_compiled.append(re.compile(regex))

        self.application_ids = self.getlist('filter', 'application_ids', [])
        self.start_time = self.get('filter', 'start_time', None)
        self.end_time = self.get('filter', 'end_time', None)

        self.namespace = self.get('options', 'namespace', 'appd')
        self.min_delay = int(self.get('options', 'min_delay', 60))
        self.skip_null_values = self.getboolean(
            'options', 'skip_null_values', False)
        self.default_null_value = self.get('options', 'default_null_value', 0)

        self._setup_output(self)
        self.last_run_time = self.get_last_run_time()
        if self.start_time and self.end_time and self.last_run_time:
            if self.last_run_time > self.start_time:
                self.start_time = self.last_run_time
        elif self.last_run_time:
            self.start_time = self.last_run_time

        self.writer_host = self.get('writer', 'host', '127.0.0.1')
        self.writer_port = int(self.get('writer', 'port', '2878'))
        self.is_dry_run = self.getboolean('writer', 'dry_run', True)

    def get_value_to_send(self, name, value):
        """
        Gets the value to send for the given value.  This allows us to translate
        the value from "NaN" to "0", for example.
        This function returns None when it should not be sent.
        Arguments:
        name - the name of the metric
        value - the value to check

        Return:
        The value to send or None to not send
        """

        if (not value or
                str(value).lower() == 'nan' or
                not isinstance(value, numbers.Number)):
            value = self.default_null_value

        return value

    def validate(self):
        """
        Checks that all required configuration items are set
        Throws:
        ValueError when a configuration item is missing a value
        """

        if not self.api_username:
            raise ValueError('api.username configuration is required')
        if not self.api_password:
            raise ValueError('api.password configuration is required')
        if not self.api_account:
            raise ValueError('api.account configuration is required')
        if not self.api_url:
            raise ValueError('api.controller_url configuration is required')

class AppDMetricRetrieverCommand(command.Command):
    """
    Command object for retrieving AppD metrics via the REST API.
    """

    def __init__(self, **kwargs):
        super(AppDMetricRetrieverCommand, self).__init__(**kwargs)
        self.description = 'AppDynamics Metric Retriever'
        self.appd_client = None
        self.config = None
        self.proxy = None

    #pylint: disable=too-many-arguments
    #pylint: disable=bare-except
    def send_metric(self, name, value, host, timestamp, tags=None,
                    value_translator=None):
        """
        Sends the metric to writer.

        Arguments:
        name - the metric name
        value - the numeric value
        host - the source/host
        timestamp - the timestamp (epoch seconds) or datetime object
        tags - dictionary of tags
        value_translator - function pointer to function that will translate
            value from current form to something else
        """

        if not isinstance(timestamp, numbers.Number):
            parsed_date = datetime.datetime.strptime(timestamp,
                                                     '%Y-%m-%dT%H:%M:%S+00:00')
            parsed_date = parsed_date.replace(tzinfo=dateutil.tz.tzutc())
            timestamp = utils.unix_time_seconds(parsed_date)

        if value_translator:
            value = value_translator(name, value)
            if value is None:
                return

        attempts = 0
        while attempts < 5 and not utils.CANCEL_WORKERS_EVENT.is_set():
            try:
                self.proxy.transmit_metric(self.config.namespace + '.' +
                                           utils.sanitize_name(name),
                                           value, int(timestamp), host, tags)
                break
            except:
                attempts = attempts + 1
                self.logger.warning('Failed to transmit metric %s: %s',
                                    name, str(sys.exc_info()))
                if not utils.CANCEL_WORKERS_EVENT.is_set():
                    time.sleep(1)

    #pylint: disable=no-self-use
    def get_help_text(self):
        """
        Help text for this command.
        """

        return "Pull metrics from AppDynamics"

    def _initialize(self, arg):
        """
        Parses the arguments passed into this command.

        Arguments:
        arg - the argparse parser object returned from argparser
        """

        self.config = AppDPluginConfiguration(arg.config_file_path)
        self.config.validate()
        try:
            logging.config.fileConfig(arg.config_file_path)
        except ConfigParser.NoSectionError:
            pass
        self.logger = logging.getLogger()

    #pylint: disable=too-many-branches
    def _execute(self):
        """
        Execute this command
        """

        # connect to the wf proxy
        self.proxy = WavefrontMetricsWriter(self.config.writer_host,
                                            self.config.writer_port,
                                            self.config.is_dry_run)
        self.proxy.start()

        # connect to appd
        self.appd_client = AppDynamicsClient(self.config.api_url,
                                             self.config.api_username,
                                             self.config.api_password,
                                             self.config.api_account,
                                             self.config.api_debug)

        # construct start time for when to get metrics starting from
        if self.config.start_time:
            start = (dateutil.parser.parse(self.config.start_time)
                     .replace(microsecond=0, tzinfo=dateutil.tz.tzutc()))
        else:
            start = ((datetime.datetime.utcnow() -
                      datetime.timedelta(seconds=60.0))
                     .replace(microsecond=0, tzinfo=dateutil.tz.tzutc()))
        if self.config.end_time:
            end = (dateutil.parser.parse(self.config.end_time)
                   .replace(microsecond=0, tzinfo=dateutil.tz.tzutc()))
        else:
            end = None

        if start is not None:
            if end is None:
                end = (datetime.datetime.utcnow()
                       .replace(microsecond=0, tzinfo=dateutil.tz.tzutc()))

            if (end - start).total_seconds() < self.config.min_delay:
                self.logger.info('Not running since %s - %s < 60',
                                 str(end), str(start))
                return

        start = start.replace(microsecond=0, tzinfo=dateutil.tz.tzutc())
        self.logger.info('Running %s - %s', str(start), str(end))

        for app in self.appd_client.get_applications():
            if str(app.id) not in self.config.application_ids:
                print 'skipping %s (%s)' % (app.name, str(app.id))
                continue
            if utils.CANCEL_WORKERS_EVENT.is_set():
                break

            # get a list of metrics available
            # TODO: cache this like New Relic plugin
            self.logger.info('[%s] Getting metric tree', app.name)
            paths = self.get_metric_paths(app)

            # if the time is more than 10 minutes, AppD will make sample size
            # larger than a minute.  so, we'll grab the data in chunks
            # (10m at a time)
            curr_start = start
            if not end:
                end = (datetime.datetime.utcnow()
                       .replace(microsecond=0, tzinfo=dateutil.tz.tzutc()))
            curr_end = end
            curr_diff = curr_end - curr_start
            while (curr_diff.total_seconds() > 0 and
                   not utils.CANCEL_WORKERS_EVENT.is_set()):
                if (curr_diff.total_seconds() > 600 or
                        curr_diff.total_seconds() < 60):
                    curr_end = curr_start + datetime.timedelta(minutes=10)

                self._process_metrics(paths, app, curr_start, curr_end)

                # save "last run time" and update curr_* variables
                self.config.set_last_run_time(curr_end)
                curr_start = curr_end
                curr_end = end
                curr_diff = curr_end - curr_start
                if (curr_diff.total_seconds() > 600 and
                        not utils.CANCEL_WORKERS_EVENT.is_set()):
                    time.sleep(30)

    def get_metric_paths(self, app):
        """
        Calls the get_metric_tree() api for the given app and returns
        all paths that are not in the black list (or are in black but included
        in white list)
        Arguments:
        app - the application object
        See:
        _get_metric_paths()
        """
        metric_tree = self.appd_client.get_metric_tree(app.id, None, True)
        paths = []
        self._get_metric_paths(paths, app, metric_tree)
        return paths

    def _get_metric_paths(self, _rtn_paths, app, metric_tree):
        """
        Gets a list of paths to retrieve from get_metrics()
        Arguments:
        _rtn_paths: out argument to return the list of paths (for recursion)
        app: the application object
        metric_tree: the response from get_metric_tree()
        """

        for node in metric_tree:
            if utils.CANCEL_WORKERS_EVENT.is_set():
                break

            if node.type == 'folder':
                self._get_metric_paths(_rtn_paths, app, node._children)
                continue

            # black list ...
            keep = True
            for pattern in self.config.fields_blacklist_regex_compiled:
                keep = not pattern.match(node.path)

            # white list ...
            if not keep:
                for pattern in self.config.fields_whitelist_regex_compiled:
                    keep = pattern.match(node.path)

            if keep:
                _rtn_paths.append(node.path)

    def _process_metrics(self, paths, app, start, end):
        """
        Processes metrics returned from a get_metrics() api call.
        Arguments:
        paths - list of paths returned from get_metric_paths()
        app - the application object
        start - the start datetime object
        end - the end datetime object
        """

        for path in paths:
            if utils.CANCEL_WORKERS_EVENT.is_set():
                break
            self.logger.info('[%s] Getting \'%s\' metrics for %s - %s',
                             app.name, path, start, end)
            metrics = self.appd_client.get_metrics(
                path, app.id, 'BETWEEN_TIMES', None,
                long(utils.unix_time_seconds(start) * 1000),
                long(utils.unix_time_seconds(end) * 1000),
                False)
            for metric in metrics:
                if utils.CANCEL_WORKERS_EVENT.is_set():
                    break
                for value in metric.values:
                    self.send_metric(self.config.namespace + '|' + path,
                                     value.current,
                                     'appd', # the source name
                                     long(value.start_time_ms / 1000),
                                     None, # tags
                                     self.config.get_value_to_send)



"""
Command base class and utility functions
"""

import datetime
import gc
import logging
import os.path
import time

import dateutil.tz

from wavefront import utils

#pylint: disable=unused-argument
#pylint: disable=no-self-use
class Command(object):
    """
    This is the base class of each command implemented.
    """

    def __init__(self, **kwargs):
        super(Command, self).__init__()

        self.verbose = False
        self.logger = logging.getLogger()
        if len(self.logger.root.handlers) > 0:
            fmt = logging.Formatter('%(levelname)s: %(thread)d %(message)s')
            self.logger.root.handlers[0].setFormatter(fmt)
        self.description = kwargs.get('description', 'Wavefront command')
        self.name = kwargs.get('name', 'wfcollector')

    def _initialize(self, args):
        """
        Initializes the command

        Arguments:
        args - list of arguments
        """

        raise ValueError('command:_initiailize() should be implemented '
                         'by subclass')

    def add_arguments(self, parser):
        """
        Adds this command's arguments to the argsparser object

        Arguments:
        parser - the subparser where arguments are added
        """

        default_config = '/opt/wavefront/etc/' + self.name + '.conf'
        parser.add_argument('--config',
                            dest='config_file_path',
                            default=default_config,
                            help='Path to configuration file')

    #pylint: disable=bare-except
    def execute(self, args):
        """
        Execute this command with the given arguments.

        Arguments:
        arg - the argparse parser object returned from argparser
        """

        self._initialize(args)
        while not utils.CANCEL_WORKERS_EVENT.is_set():
            try:
                self.logger.info('Executing %s ...', self.description)
                self._execute()

            except:
                self.logger.exception('Failed executing %s', self.description)

            if utils.CANCEL_WORKERS_EVENT.is_set():
                break

            if 'delay' not in args or not args.delay or args.delay <= 0:
                break

            gc.collect()
            if not utils.CANCEL_WORKERS_EVENT.is_set():
                self.logger.info('Sleeping for %ds ...', args.delay)
                time.sleep(args.delay)

    def _execute(self):
        """
        Override this in the subclass to perform the command.
        """
        raise ValueError('command:_execute() should be implemented by subclass')

    def get_help_text(self):
        """
        Gets the text to display for this command.
        """

        return ""

# pylint: disable=too-few-public-methods
class CommandConfiguration(utils.Configuration):
    """
    Base class for configurations of a command
    """

    def __init__(self, config_file_path, create_if_not_exist=False):
        super(CommandConfiguration, self).__init__(
            config_file_path=config_file_path,
            create_if_not_exist=create_if_not_exist)

        self.last_run_time_section = 'options'
        self.output_directory = None
        self.output = None

    def _setup_output(self, config):
        """
        Sets up the output directory and output file
        """

        self.output_directory = config.get('options', 'output_directory', None)

        if self.output_directory:
            if not os.path.exists(self.output_directory):
                os.makedirs(self.output_directory)

            output_file = (self.output_directory + '/' +
                           os.path.basename(self.config_file_path) + '.save')
        else:
            output_file = ('/tmp/' +
                           os.path.basename(self.config_file_path) + '.save')

        # try to touch the file to see if we have permission
        try:
            with open(output_file, 'a'):
                os.utime(output_file, None)
        except IOError:
            raise ValueError('Unable to write to output file ' + output_file)

        self.output = utils.Configuration(
            config_file_path=output_file,
            create_if_not_exist=True)

    def get_last_run_time(self, section_name=None):
        """
        Gets the last run time as a string
        """

        if section_name:
            return self.output.getdate(section_name, 'last_run_time', None)

        else:
            return self.output.getdate(
                self.last_run_time_section, 'last_run_time', None)

    def set_last_run_time(self, run_time, section_name=None):
        """
        Sets the last run time to the run_time argument.

        Arguments:
        run_time - the time when this script last executed successfully (end)
        """

        if utils.CANCEL_WORKERS_EVENT.is_set():
            return

        if not run_time:
            run_time = (datetime.datetime.utcnow()
                        .replace(microsecond=0, tzinfo=dateutil.tz.tzutc()))
        if section_name:
            self.output.set(section_name, 'last_run_time',
                            run_time.isoformat())
        else:
            self.output.set(self.last_run_time_section, 'last_run_time',
                            run_time.isoformat())
        self.output.save()



#!/usr/bin/env python
"""
This is the system checker plugin for Wavefront.  It runs as part of the
wavefront integrations tool package.  It should be run on each host/system
where you want to check for things like core files, etc
"""

import ConfigParser
import fnmatch
import hashlib
import httplib
import logging
import os
import os.path
import socket
import sys
import time

import wavefront_client
from wavefront_client.rest import ApiException
from wavefront.utils import Configuration
from wavefront import command
from wavefront import utils

# default location for the configuration file.
DEFAULT_CONFIG_FILE_PATH = '/opt/wavefront/etc/system_checker.conf'

#pylint: disable=too-many-instance-attributes
class SystemCheckerConfiguration(Configuration):
    """
    Configuration interface for system checker
    """

    def __init__(self, config_file_path):
        super(SystemCheckerConfiguration, self).__init__(
            config_file_path=config_file_path)

        # section: global
        self.cache_location = self.get('global', 'cache_dir', '/tmp')
        self.source_name = self.get('global', 'source_name',
                                    socket.gethostname())
        self.log_requests = self.getboolean('global', 'log_requests', False)
        self.ignore_ssl_cert_errors = self.getboolean(
            'global', 'ignore_ssl_cert_errors', False)

        # section: wavefront
        self.wf_api_key = self.get('wavefront', 'api_key', None)
        self.wf_api_base = self.get('wavefront', 'api_base',
                                    'https://metrics.wavefront.com')

        # section: find_files
        self.find_file_locations = self.getlist('find_files', 'paths', [])
        self.find_file_patterns = self.getlist('find_files', 'patterns', [])
        self.find_file_event_names = self.getlist(
            'find_files', 'event_names', [])

        # section: files_changed
        self.file_changed_files = self.getlist('file_changes', 'files', [])
        self.file_changed_event_names = self.getlist(
            'file_changes', 'event_names', [])

        # other instance variables (non-configuration)
        self.md5_config = None
        self.md5_hashes = None

        # initialize the cache directory
        self._init_cache()

    def _init_cache(self):
        """
        Initializes the cache directory
        """

        if not os.path.exists(self.cache_location):
            os.mkdir(self.cache_location)
            os.mkdir(os.path.join(self.cache_location, 'file-changes'))
            os.mkdir(os.path.join(self.cache_location, 'find-files'))

        self.md5_config = utils.Configuration(os.path.join(
            self.cache_location, 'file-changes', 'wf_md5_hashes.conf'), True)
        self.md5_hashes = {}
        if self.md5_config.has_section('hashes'):
            items = self.md5_config.config.items('hashes')
            for item in items:
                self.md5_hashes[item[0]] = item[1]

        # create directory to store hashes of files found
        for path in self.find_file_locations:
            cache_path = os.path.join(
                self.cache_location, 'find-files', utils.sanitize_name(path))
            if not os.path.exists(cache_path):
                os.makedirs(cache_path)

    def _get_file_found_cache_path(self, directory, filename, hashval):
        """
        Gets the file for a specific file and hash value combo for the cache
        directory.
        Arguments:
        directory - the directory where the file was found
        filename - the file name of the file found
        hashval - the md5 hash value of the file found
        """

        cache_path = os.path.join(
            self.cache_location, 'find-files', utils.sanitize_name(directory))
        sfilename = utils.sanitize_name(filename) + '_' + hashval
        return os.path.join(cache_path, sfilename)

    def has_file_found_cache_path(self, directory, filename, hashval):
        """
        Checks to see if the given file has already been reported.
        Arguments:
        directory - the directory where the file was found
        filename - the file name of the file found
        hashval - the md5 hash value of the file found
        """

        final_path = self._get_file_found_cache_path(
            directory, filename, hashval)
        return os.path.exists(final_path)

    def set_file_found(self, directory, filename, hashval):
        """
        Sets the file was found and its hash value.
        Arguments:
        directory - the directory where the file was found
        filename - the file name to the file found
        hashval - the md5 hash value of the file found
        """

        final_path = self._get_file_found_cache_path(
            directory, filename, hashval)
        with open(final_path, 'a'):
            os.utime(final_path, None)

    def validate(self):
        """
        Validates the configuration values
        Throws:
        ValueError when length of find files locations array does not equal
                   length of patterns and event names arrays
        """

        if len(self.find_file_locations) != len(self.find_file_patterns):
            raise ValueError('find_files:paths must have the same number of '
                             'elements as find_files:patterns')
        if len(self.find_file_locations) != len(self.find_file_event_names):
            raise ValueError('find_files:paths must have the same number of '
                             'elements as find_files:event_names')

    def set_expected_hash(self, filename, hashval):
        """
        Sets the expected hash value for the given index

        Arguments:
        filename - the name of the file
        hashval - the md5 hash value to update to
        """

        self.md5_hashes[filename] = hashval
        self.md5_config.set('hashes', filename, hashval)
        self.md5_config.save()

class SystemCheckerCommand(command.Command):
    """
    System checker command class
    """

    def __init__(self, **kwargs):
        super(SystemCheckerCommand, self).__init__(**kwargs)
        self.config = None
        self.description = "System Checker"

    #pylint: disable=no-self-use
    def get_help_text(self):
        """
        Help text for this command.
        """

        return ('System Checker for finding files matching a pattern, '
                'files that changed, etc.')

    def _initialize(self, arg):
        """
        Parses the arguments passed into this command.

        Arguments:
        arg - the argparse parser object returned from argparser

        Raises:
        ValueError - when config file is not provided
        """

        if 'config_file_path' not in arg:
            raise ValueError('--config parameter is required')

        self.config = SystemCheckerConfiguration(arg.config_file_path)
        self.config.validate()
        try:
            logging.config.fileConfig(arg.config_file_path)
        except ConfigParser.NoSectionError:
            pass

        # logger
        self.logger = logging.getLogger()
        if self.config.log_requests:
            httplib.HTTPConnection.debuglevel = 1

        # configure wavefront api
        wavefront_client.configuration.api_key['X-AUTH-TOKEN'] = \
          self.config.wf_api_key
        wavefront_client.configuration.host = self.config.wf_api_base
        wavefront_client.configuration.verify_ssl = (
            not self.config.ignore_ssl_cert_errors)

    #pylint: disable=bare-except
    #pylint: disable=too-many-arguments
    def _send_event(self, name, description, start, end, severity, etype):
        """
        Sends event to wavefront API

        Arguments:
        name - event name
        description -
        start -
        end -
        severity -
        etype - event type

        Returns:
        True if successfully created event, false o/w
        """

        events_api = wavefront_client.EventsApi()
        attempts = 0
        sleep_time = 1
        successful = False
        while attempts < 5 and not utils.CANCEL_WORKERS_EVENT.is_set():
            self.logger.info('%s Creating event %s', self.description, name)
            try:
                if start == end:
                    events_api.create_new_event(
                        name,
                        s=int(start),
                        c=True,
                        d=description,
                        h=[self.config.source_name, ],
                        l=severity,
                        t=etype)
                else:
                    events_api.create_new_event(
                        name,
                        s=int(start),
                        e=int(end),
                        c=False,
                        d=description,
                        h=[self.config.source_name, ],
                        l=severity,
                        t=etype)
                successful = True
                break

            except ApiException as api_ex:
                self.logger.warning('Failed to send event: %s (attempt %d)\n%s',
                                    api_ex.reason, attempts+1, api_ex.body)

            except:
                self.logger.warning('Failed to send event: %s (attempt %d)',
                                    str(sys.exc_info()), attempts+1)

            if not successful:
                attempts = attempts + 1
                if not utils.CANCEL_WORKERS_EVENT.is_set():
                    time.sleep(sleep_time)
                    sleep_time = sleep_time * 2

        return successful

    def _check_for_files_matching(self):
        """
        Checks for files matching a pattern in the configured paths
        """

        self.logger.info('Checking files in ' +
                         str(self.config.find_file_locations))
        for path in self.config.find_file_locations:
            self.logger.info('Looking for matching files in %s ...', path)
            if not os.path.exists(path):
                self.logger.warning('Path %s does not exist.', path)
                continue

            for filename in os.listdir(path):
                loc = 0
                for pattern in self.config.find_file_patterns:
                    if pattern and fnmatch.fnmatch(filename, pattern):
                        fullpath = os.path.join(path, filename)
                        hashval = utils.hashfile(fullpath, hashlib.md5())
                        if self.config.has_file_found_cache_path(
                                path, filename, hashval):
                            continue
                        self.logger.warning('[%s: %s] Found file "%s" matching '
                                            'pattern "%s"', self.description,
                                            path, filename, pattern)
                        created = os.path.getctime(fullpath)
                        name = self.config.find_file_event_names[loc]
                        if self._send_event(
                                name + ' found',
                                name + ' file found: ' + fullpath,
                                created,
                                created,
                                'Warning',
                                name):
                            self.config.set_file_found(path, filename, hashval)

                        loc = loc + 1

    def _check_for_files_changed(self):
        """
        Checks the hash (md5 currently) for each file configured
        """

        loc = 0
        for path in self.config.file_changed_files:
            try:
                self._check_file_hash(
                    path, self.config.file_changed_event_names[loc])
            except IOError as ioe:
                self.logger.error('Unable to check MD5 for %s: %s',
                                  path, str(ioe))
            loc = loc + 1

    def _check_file_hash(self, path, event_name):
        """
        Check a specific path's hash against its expected value
        Arguments:
        path - the full path to the file to check
        event_name - the event name to use when creating an event (on hash chng)
        """

        self.logger.info('Checking MD5 for %s ...', path)
        hashval = utils.hashfile(path, hashlib.md5())
        abspath = os.path.abspath(path)
        if (abspath not in self.config.md5_hashes or
                not self.config.md5_hashes[abspath]):
            # assume this is the first run
            self.config.set_expected_hash(abspath, hashval)

        else:
            expected_hashval = self.config.md5_hashes[abspath]

            if expected_hashval != hashval:
                modified = os.path.getmtime(path) * 1000
                self.logger.warning('[%s: %s] MD5 mismatch. '
                                    'Expected: %s; Found: %s',
                                    self.description, path, expected_hashval,
                                    hashval)
                self._send_event('File Change (' + path + ')',
                                 'File Change (' + path + ')',
                                 modified,
                                 modified,
                                 'Informational',
                                 event_name)

                # update the new expected hash to this value
                self.config.set_expected_hash(abspath, hashval)

    def _execute(self):
        """
        Starts looking for matching files, changed files, etc as configured
        """

        if not utils.CANCEL_WORKERS_EVENT.is_set():
            self._check_for_files_matching()
        if not utils.CANCEL_WORKERS_EVENT.is_set():
            self._check_for_files_changed()



"""
This is common code used by the New Relic plugin.
"""

import datetime
import json
import logging
import numbers
import sys
import time

import urllib2
import urlparse
import dateutil

from wavefront.metrics_writer import WavefrontMetricsWriter
from wavefront import command
from wavefront import utils

# http://bugs.python.org/issue7980
# bug indicates that datetime.datetime.strptime() must be called once before
# multiple threads are started that use this function.
datetime.datetime.strptime('2016-04-27T13:30:15+00:00',
                           '%Y-%m-%dT%H:%M:%S+00:00')

class NewRelicCommand(command.Command):
    """
    Base class for all New Relic command objects
    """

    def __init__(self, **kwargs):
        super(NewRelicCommand, self).__init__(**kwargs)
        self.config = None
        self.proxy = None

    def init_proxy(self):
        """
        Initializes the proxy writer
        """

        self.proxy = self.get_writer_from_config(self.config)

    @staticmethod
    def get_writer_from_config(config):
        """
        Creates a new metrics writer pointed to the proxy using the given
        config object and starts it

        Arguments:
        config - the configuration
        """
        proxy = WavefrontMetricsWriter(config.writer_host,
                                       config.writer_port,
                                       config.is_dry_run)
        proxy.start()
        return proxy

    #pylint: disable=too-many-arguments
    #pylint: disable=bare-except
    @staticmethod
    def send_metric(writer, name, value, host, timestamp, tags=None,
                    value_translator=None, logger=None):
        """
        Sends the metric to writer.

        Arguments:
        name - the metric name
        value - the numeric value
        host - the source/host
        timestamp - the timestamp (epoch seconds) or datetime object
        tags - dictionary of tags
        value_translator - function pointer to function that will translate
            value from current form to something else
        """

        if not isinstance(timestamp, numbers.Number):
            parsed_date = datetime.datetime.strptime(timestamp,
                                                     '%Y-%m-%dT%H:%M:%S+00:00')
            parsed_date = parsed_date.replace(tzinfo=dateutil.tz.tzutc())
            timestamp = utils.unix_time_seconds(parsed_date)

        if value_translator:
            value = value_translator(name, value)
            if value is None:
                return

        attempts = 0
        while attempts < 5 and not utils.CANCEL_WORKERS_EVENT.is_set():
            try:
                writer.transmit_metric('newrelic.' + utils.sanitize_name(name),
                                       value, int(timestamp), host, tags)
                break
            except:
                attempts = attempts + 1
                logger.warning('Failed to transmit metric %s: %s',
                               name, str(sys.exc_info()))
                if not utils.CANCEL_WORKERS_EVENT.is_set():
                    time.sleep(1)

    #pylint: disable=line-too-long
    def call_paginated_api(self, path, query_string, callback, callback_args):
        """
        Calls the New Relic API that contains the pagination information
        in the response and loops over all pages returning a structured
        response with each page having its own key.
        Expects to find a Link header in the response that has page details
        contained within it.  Something like this:
        HTTP/1.1 200 OK
        Link: <https://api.newrelic.com/v2/applications/13812111/metrics.json?page=1>; rel="first", <https://api.newrelic.com/v2/applications/13812111/metrics.json?page=1>; rel="prev", <https://api.newrelic.com/v2/applications/13812111/metrics.json?page=3>; rel="next", <https://api.newrelic.com/v2/applications/13812111/metrics.json?page=23>; rel="last"

        Arguments:
        path - the path of the API
        query_string - the query string (optional)
        callback - the function handling each page's response
        callback_args - the args to pass to the page callback function
        """

        full_response = []
        def get_page_response(page):
            """
            Gets a single page's response and calls the callback function
            Arguments:
            page - the page to retrieve
            Returns:
            The tuple of (json response, original urllib2 response object)
            """

            if query_string:
                page_query_string = query_string.copy()
            else:
                page_query_string = {}
            page_query_string['page'] = page

            tpl = self.call_api(path, page_query_string)
            if utils.CANCEL_WORKERS_EVENT.is_set():
                return tpl

            if callback:
                args = (tpl[0], ) + callback_args
                callback(*args)
            else:
                full_response.append(tpl[0])
            return tpl

        # get the first page and see if there are more by inspecting header
        tpl = get_page_response(1)
        if utils.CANCEL_WORKERS_EVENT.is_set():
            return full_response

        link = self.parse_link_header(tpl[1].info().getheader('Link'))

        # build a list of function pointers - one per page
        function_pointers = []
        for page in range(2, link['last'] + 1):
            function_pointers.append((get_page_response, (page,)))

        utils.parallel_process_and_wait(function_pointers, self.config.workers,
                                        self.logger)
        # empty unless callback is None
        return full_response

    def call_api(self, path, query_string=None):
        """
        Calls the New Relic API in a retry loop until no error or limit
        is reached

        Arguments:
        path - the path of the API
        query_string - the query string (optional)

        Returns:
        Tuple: JSON object parsed from the string returned by the URL and the
        original HTTP response object.
        """

        attempts = 0
        sleep_time = 5
        while attempts < 5 and not utils.CANCEL_WORKERS_EVENT.is_set():
            try:
                return self._call_api(path, query_string)
            except urllib2.HTTPError, err:
                if err.code == 500:
                    attempts = attempts + 1
                    if not utils.CANCEL_WORKERS_EVENT.is_set():
                        time.sleep(sleep_time)
                        sleep_time = sleep_time * 2
                    else:
                        return (None, None)
                else:
                    raise

            except ValueError:
                attempts = attempts + 1
                if not utils.CANCEL_WORKERS_EVENT.is_set():
                    time.sleep(sleep_time)
                    sleep_time = sleep_time * 2
                else:
                    return (None, None)

        return (None, None)


    def _call_api(self, path, query_string=None):
        """
        Calls the New Relic API and parses response as JSON

        Arguments:
        path - the path of the API
        query_string - the query string (optional)

        Returns:
        Tuple: JSON object parsed from the string returned by the URL and the
        original HTTP response object.
        """

        # build the URL
        url = self.config.api_endpoint + path
        self.logger.debug('%s', url)
        if query_string:
            url = '%s?%s' % (url, utils.urlencode_utf8(query_string))

        if self.config.api_log_path:
            with open(self.config.api_log_path, 'a') as log_fd:
                log_fd.write(url + '\n')

        # create the request object
        req = urllib2.Request(url=url)
        req.add_header('Accept', 'application/json')
        req.add_header(self.config.api_key_header_name, self.config.api_key)

        try:
            response = urllib2.urlopen(req, timeout=30)
            json_response = json.load(response)

        except urllib2.HTTPError as http_err:
            self.logger.warning('Failed [%s]: %s', url, str(http_err))
            try:
                json_response = json.load(http_err)

            except:
                json_err = sys.exc_info()[1]
                self.logger.warning('Failed to load JSON [%s]: %s',
                                    url, str(json_err))
                raise ValueError(str(json_err))

        except urllib2.URLError as url_err:
            self.logger.warning('Failed [%s]: %s', url, str(url_err))
            raise ValueError(str(url_err))

        except BaseException as base_ex:
            self.logger.warning('Failed (1) [%s]: %s', url, str(base_ex))
            raise ValueError(str(base_ex))

        except:
            self.logger.warning('Failed [%s]: %s', url, str(sys.exc_info()))
            raise ValueError('Unknown failure with ' + url)

        if self.config.api_log_path:
            with open(self.config.api_log_path, 'a') as log_fd:
                log_fd.write('%s\n' % (str(json_response)))

        if 'error' in json_response:
            if 'title' in json_response['error']:
                self.logger.warning('%s\n%s', json_response['error']['title'],
                                    str(json_response))
                raise ValueError(json_response['error']['title'])
            else:
                self.logger.warning('%s\n%s', str(json_response['error']),
                                    str(json_response))
                raise ValueError(json_response['error'])

        return (json_response, response)

    @staticmethod
    def parse_link_header(link):
        """
        Parses the Link header and returns a dictionary with keys:
           first: the first page
           next: the next page
           last: the last page
           previous: the previous page

        All will be 0 if not set

        Arguments:
        link - the link header string

        Returns:
        dictionary
        """

        rtn = {
            'first': 0,
            'next': 0,
            'last': 0,
            'previous': 0
        }

        if not link:
            return rtn

        #  Link: <.../v2/applications/X/metrics.json?page=1>; rel="first",
        parts = link.split(', ')
        for part in parts:
            components = part.split('; ')
            url = components[0][1:-1]
            result = urlparse.parse_qs(urlparse.urlparse(url).query)
            page = int(result['page'][0])
            if components[1] == 'rel="first"':
                rtn['first'] = page
            elif components[1] == 'rel="prev"':
                rtn['previous'] = page
            elif components[1] == 'rel="last"':
                rtn['last'] = page
            elif components[1] == 'rel="next"':
                rtn['next'] = page

        return rtn
